//////////////////////////////////////////////
//                                          //
// project  : mVISIO v1.6a                  //
// file     : mvisio_terrain.H	    		  //
// author   : Achille Peternier             // 
//          : Sofien Bouaziz                //
// copyright: EPFL (C) 2005-2008            //
//												        //
//////////////////////////////////////////////

#include <fstream>

/////////////
// #DEFINE //
/////////////

   // Offset macros:
   #define MV_TERRAIN_CHUNK_BUFFEROFFSET(i) ((char*)NULL + (i))
   #define MV_TERRAIN_CHUNK_VERTEXOFFSET 0
   #define MV_TERRAIN_CHUNK_NORMALOFFSET (3*sizeof(float))
   #define MV_TERRAIN_CHUNK_TANGENTOFFSET (6*sizeof(float)) 

   // Border kind to avoid cracks:
   enum BORDER_TYPE
   {
	   NORMAL,
	   LEFT,
	   RIGHT,
	   LEFT_RIGHT
   };
   


/////////////
// STRUCTS //
/////////////

typedef struct MVTERRAIN_TERRAININFO
{
	unsigned short chunkSize;
	unsigned short height;
	unsigned short width;
	float offsetXY;
	float offsetZ;
	float elevation;
	unsigned char nbLOD;
} MVTERRAIN_TERRAININFO;

typedef struct MVTERRAIN_CHUNKINFO
{
	unsigned short x;
	unsigned short y;
	float hmin;
	float hmax;
	unsigned short top;
	unsigned short bottom;
	unsigned short right;
	unsigned short left;
} MVTERRAIN_CHUNKINFO;
   



////////////////////////////
// CLASS MVTERRAIN_VERTEX //
////////////////////////////

///////////////////////////////////////////////////
typedef class MVTERRAIN_VERTEX // Terrain vertex //
///////////////////////////////////////////////////
{
/////////////
   public: //
/////////////
      
   // Const/dest:   
	inline void setPosition(float x, float y, float z)	  { position.set(x, y, z); }
	inline void setNormal(float nX, float nY, float nZ)  { normal.set(nX, nY, nZ); }	
	inline void setTangent(float tX, float tY, float tZ) { tangent.set(tX, tY, tZ); }				
		
		
//////////////
   private: //
//////////////
	
	// Data:
	MVVECTOR3 position, normal, tangent;		
} MVTERRAIN_VERTEX;



/////////////////////////////////
// CLASS MVTERRAIN_INDEXBUFFER //
/////////////////////////////////

//////////////////////////////////////////////////////////////
typedef class MVTERRAIN_INDEXBUFFER // Terrain index buffer //
//////////////////////////////////////////////////////////////
{
/////////////
   public: //
/////////////
	
	// Const/dest:
	MVTERRAIN_INDEXBUFFER(unsigned int _num, unsigned long _nbIndices)
	{
		num = _num;
		nbIndices = _nbIndices;
	}
	inline unsigned int *getBufferNumber()     { return &num; }
	inline unsigned int getNumberOfIndices()  { return nbIndices; }
	
	
//////////////
   private: //
//////////////	

   // Data:
	unsigned int num;
	unsigned int nbIndices;
} MVTERRAIN_INDEXBUFFER;



/////////////////////////////////
// CLASS MVTERRAIN_COORDBUFFER //
/////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
typedef class MV_API MVTERRAIN_COORDBUFFER // Terrain texture coord buffer //
/////////////////////////////////////////////////////////////////////////////
{
/////////////
   public: //
/////////////

	// Const/dest:
	MVTERRAIN_COORDBUFFER(MVUV* _textureUV, unsigned short _chunkSize) 
	{		
	   MVOPENGL::glGenBuffersARB(1, &num);
	   MVOPENGL::glBindBufferARB(GL_ARRAY_BUFFER_ARB, num);					
	   MVOPENGL::glBufferDataARB(GL_ARRAY_BUFFER_ARB, 
							           (_chunkSize)*(_chunkSize)*sizeof(MVUV),
							            _textureUV, 
							            GL_STATIC_DRAW_ARB);
	} 	
	~MVTERRAIN_COORDBUFFER()
	{
	   MVOPENGL::glDeleteBuffersARB(1, &num);
	}
	
	// Settings:
	inline unsigned int getBuffer() { return num; }


//////////////
   private: //
//////////////  
	
	// Data:
	unsigned int num;
} MVTERRAIN_TEXTURE;



///////////////////////////
// CLASS MVTERRAIN_INDEX //
///////////////////////////

/////////////////////////////////////////////////
typedef class MVTERRAIN_INDEX // Terrain index //
/////////////////////////////////////////////////
{   	
/////////////
	public: //	
/////////////

   // Const/dest:
   MVTERRAIN_INDEX(unsigned short _chunkSize, unsigned char _nbLevels);
	~MVTERRAIN_INDEX();

	inline unsigned char getNbLevels(void) { return nbLevels; }

	inline MVTERRAIN_INDEXBUFFER* getCenterIndexBuffer(unsigned char _level)                        { return centerIndexBuffers[_level]; }
	inline MVTERRAIN_INDEXBUFFER* getLowTopIndexBuffer(unsigned char _level, BORDER_TYPE _type)     { return lowTopIndexBuffers[_level*4 + _type]; }
	inline MVTERRAIN_INDEXBUFFER* getLowBottomIndexBuffer(unsigned char _level, BORDER_TYPE _type)  { return lowBottomIndexBuffers[_level*4 + _type]; }
	inline MVTERRAIN_INDEXBUFFER* getLowRightIndexBuffer(unsigned char _level, BORDER_TYPE _type)   { return lowRightIndexBuffers[_level*4 + _type]; }
	inline MVTERRAIN_INDEXBUFFER* getLowLeftIndexBuffer(unsigned char _level, BORDER_TYPE _type)    { return lowLeftIndexBuffers[_level*4 + _type]; }
	inline MVTERRAIN_INDEXBUFFER* getTopIndexBuffer(unsigned char _level, BORDER_TYPE _type)        { return topIndexBuffers[_level*4 + _type]; }
	inline MVTERRAIN_INDEXBUFFER* getBottomIndexBuffer(unsigned char _level, BORDER_TYPE _type)     { return bottomIndexBuffers[_level*4 + _type]; }
	inline MVTERRAIN_INDEXBUFFER* getRightIndexBuffer(unsigned char _level, BORDER_TYPE _type)      { return rightIndexBuffers[_level*4 + _type]; }
	inline MVTERRAIN_INDEXBUFFER* getLeftIndexBuffer(unsigned char _level, BORDER_TYPE _type)       { return leftIndexBuffers[_level*4 + _type]; }

	void createCenterIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned long _nbIndices, unsigned short* _index);
	void createLowTopIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);
	void createLowBottomIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);
	void createLowRightIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);
	void createLowLeftIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);
	void createTopIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);
	void createBottomIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);
	void createRightIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);
	void createLeftIndexBuffer(unsigned char _level, unsigned short _chunkSize, unsigned char _type, unsigned long _nbIndices, unsigned short* _index);

//////////////
   private: //
//////////////   

	MVTERRAIN_INDEXBUFFER** centerIndexBuffers;
	MVTERRAIN_INDEXBUFFER** lowTopIndexBuffers;
	MVTERRAIN_INDEXBUFFER** lowBottomIndexBuffers;
	MVTERRAIN_INDEXBUFFER** lowRightIndexBuffers;
	MVTERRAIN_INDEXBUFFER** lowLeftIndexBuffers;
	MVTERRAIN_INDEXBUFFER** topIndexBuffers;
	MVTERRAIN_INDEXBUFFER** bottomIndexBuffers;
	MVTERRAIN_INDEXBUFFER** rightIndexBuffers;
	MVTERRAIN_INDEXBUFFER** leftIndexBuffers;

	unsigned char nbLevels;	
} MVTERRAIN_INDEX; // Terrain



///////////////////////////
// CLASS MVTERRAIN_CHUNK //
///////////////////////////

/////////////////////////////////////////////////
typedef class MVTERRAIN_CHUNK // Terrain index //
/////////////////////////////////////////////////
{
/////////////
	public: //
/////////////	
   
   // Const/dest:
	MVTERRAIN_CHUNK(MVTERRAIN_TERRAININFO _terrainInfo,
					    MVTERRAIN_INDEX* _staticIndex, 
					    MVTERRAIN_TEXTURE* _staticTexCoord,
					    MVTERRAIN_CHUNKINFO _chunkInfo,
					    MVTERRAIN_VERTEX* _vertex, 
					    unsigned short _place);
	~MVTERRAIN_CHUNK();
	
	// Settings:
	void loadTexture(char* _diff1, char* _norm1, int _repeat1,
					     char* _diff2, char* _norm2, int _repeat2,
					     char* _diff3, char* _norm3, int _repeat3,
					     char* _diff4, char* _norm4, int _repeat4,
					     bool _complex);
	inline void setLOD(float* _lod) { lod = _lod; }
	inline unsigned char getLevel() { return level; }
	inline bool getRenderState() { return renderChunk; }
	void getPosition(unsigned short& _posX, unsigned short& _posY) { _posX = chunkInfo.x; _posY = chunkInfo.y; }
	void render(float _inv = 1.0f);
	void update(MVCAMERA* _camera, bool _enableCLOD,  bool _inverse);
	void updateIfWater(MVCAMERA* _camera);
	void updateIfWaterReverse(MVCAMERA* _camera);
	void setNeighbors(MVTERRAIN_CHUNK* _top, MVTERRAIN_CHUNK* _bottom, MVTERRAIN_CHUNK* _right, MVTERRAIN_CHUNK* _left);
	MVTERRAIN_CHUNKINFO chunkInfo;
	MVTERRAIN_TERRAININFO terrainInfo;	

	void setPlane(MVVECTOR4* _left, MVVECTOR4* _right, MVVECTOR4* _bottom,
				     MVVECTOR4* _top, MVVECTOR4* _near, MVVECTOR4* _far)
	{
	   frustum[0] = _left;
	   frustum[1] = _right;
	   frustum[2] = _bottom;
	   frustum[3] = _top;
	   frustum[4] = _near;
	   frustum[5] = _far;
	}


//////////////
   private: //
//////////////   

   // Methods:
	bool isInFrustrum(MVCAMERA* _camera,  bool _inverse); 
	void createBuffer(MVTERRAIN_VERTEX* _vertex);
	void computeInformation();
	
	// Neighbors:	
	MVTERRAIN_CHUNK* topChunk;
	MVTERRAIN_CHUNK* bottomChunk;
	MVTERRAIN_CHUNK* rightChunk;
	MVTERRAIN_CHUNK* leftChunk;	
	MVTERRAIN_INDEXBUFFER* center;
	MVTERRAIN_INDEXBUFFER* top;
	MVTERRAIN_INDEXBUFFER* bottom;
	MVTERRAIN_INDEXBUFFER* right;
	MVTERRAIN_INDEXBUFFER* left;
	
	// Data:
	MVMATERIAL *material;	
	MVTEXTURE *diffHeightTex1, *diffHeightTex2, *diffHeightTex3, *diffHeightTex4;
	MVTEXTURE *normalGlossTex1, *normalGlossTex2, *normalGlossTex3, *normalGlossTex4;
	int repeat1, repeat2, repeat3, repeat4;	
	unsigned char level;
	unsigned int vertexBuffer;
	MVTERRAIN_INDEX* index;
	MVTERRAIN_TEXTURE* textureCoord;
	float centerX, centerY, centerZ;
	bool renderChunk, belowWater, aboveWater;
	float x, z, x1, z1;
	float* lod;
	MVVECTOR4* frustum[6];
} MVTERRAIN_CHUNK;



/////////////////////
// CLASS MVTERRAIN //
/////////////////////

////////////////////////////////////////////////////////////
typedef class MV_API MVTERRAIN : public MVNODE // Terrain //
////////////////////////////////////////////////////////////
{   	
/////////////
	public: //	
/////////////
   
   // Const/dest:
   MVTERRAIN();
	~MVTERRAIN();
	bool load(char* _confName);
	bool render(void* _data = NULL);
	void setWaterSpeed(float _speed);
	void setSkySpeed(float _speed);
   MVSKYDOME *getSkydome() const { return skydome; }


///////////
private: //
///////////
	
	bool loadGeometry(char* _fileName);
	bool loadConfig(char* _fileName);
	void renderTerrain();
	void renderTerrainInv();
	void createTerrain();
	void loadIndex();
	void loadTextureUV();
	void loadVertices();
	void createSky();
	void createWater();
	double plane[4];
	double plane2[4];
	unsigned int fb, rb, fb2, rb2;
	MVCAMERA* camera;
	std::ifstream stream;
	MVTERRAIN_TERRAININFO terrain;
	MVTERRAIN_CHUNK** chunks;
	MVTERRAIN_INDEX* staticIndex;
	MVTERRAIN_TEXTURE* staticTextureCoord;
	MVSKYDOME* skydome;
	MVWATER* water;
	unsigned long nbChunks;
	float* lod;
	bool activeLod;
	float oldModl[16];

	MVVECTOR4 leftPlane, rightPlane, bottomPlane, topPlane, nearPlane, farPlane;
	void computeFrustrum();
} MVTERRAIN; // Terrain
